//
//  MainView.m
//  beamermath

/*
 Created on 9.8.2018 by
 
 PD Dr. Fritz Hörmann
 Mathematisches Institut - Raum 433
 Albert-Ludwigs-Universität Freiburg
 Ernst-Zermelo-Strasse 1
 D-79104 Freiburg
 eMail: fritz.hoermann@math.uni-freiburg.de
 Tel: +49 761 99520388
 
 Licensed under CC BY-SA 4.0.
 https://creativecommons.org/licenses/by-sa/4.0/legalcode
 */

/*
 // 1024 960 !
 
 #define LINEWIDTH 1.5
 #define POINTSIZE 10.0
 #define FONTSIZE 20.0
 #define FONTHEIGHT 70
 #define GLLINEWIDTH 200.0
*/

#define LINEWIDTH 3.0
#define POINTSIZE 10.0
#define FONTSIZE 30.0
#define FONTHEIGHT 70
#define GLLINEWIDTH 300.0


#import "MainView.h"
#import "Document.h"
#include <OpenGL/gl.h>
#include <sys/time.h>
#import <mach/mach_time.h>


@implementation MBBox
@end
@implementation MBPage
@end
@implementation MBCmd
@end


double condensevalue(VALUE* value, double time)
{
    if(time>value->stoptime && (value->stoptime!=0.0))
    {
        value->v += value->dv * (value->stoptime-value->time);
        value->time = value->stoptime;
        value->stoptime = 0.0; value->dv = 0.0;
    }

    if(value->time!=0.0)value->v += value->dv * (time-value->time);
    value->time = time;
    
    double res = value->v;
    
    return res;
}

void valuegoal(VALUE* value, double time, double v)
{
    if(time==0.0 || time<=value->time)
    {
        value->v = v;
        value->dv = 0.0;
        value->time = 0.0;
        value->stoptime = 0.0;
    }
    else
    {
        if(value->stoptime==0.0)
        {
            value->dv = (v-value->v)/(time-value->time);
            value->stoptime = time;
        }
    }
}

void valuesetdv(VALUE* value, double dv)
{
    value->dv = dv;
    value->stoptime = 0.0;
}

@implementation MainView


- (void)awakeFromNib
{
    ipage = 1; npage = 0;
    
    [[self window] setAcceptsMouseMovedEvents:YES];
    
    _pages = [[NSMutableDictionary alloc] init];
    _boxes = [[NSMutableDictionary alloc] init];
    
    memset(&anglex, 0, sizeof(VALUE));
    memset(&angley, 0, sizeof(VALUE));
    memset(&anglez, 0, sizeof(VALUE));
    memset(&zoom, 0, sizeof(VALUE)); zoom.v = 0.0;
    memset(&(vec[0]), 0, sizeof(VALUE));
    memset(&(vec[1]), 0, sizeof(VALUE));
    memset(&(vec[2]), 0, sizeof(VALUE));
    memset(&basisx, 0, sizeof(VALUE));
    memset(&basisy, 0, sizeof(VALUE));
    
    offset = 0.0;
    
    goalscompleted = false;
    displaymouse = true;
    
    lseries = NULL;
}

- (void)drawRect:(NSRect)dirtyRect
{
    [self draw:[[NSProcessInfo processInfo] systemUptime]];
}

void mul4(double* m, double* v)
{
    double res[4] = {0.0,0.0,0.0,0.0};
    
    for(int i=0;i<4;i++)
        for(int j=0;j<4;j++)
        {
            res[i] += m[4*j+i]*v[j];
        }
    
    for(int i=0;i<4;i++)v[i] = res[i];
}

- (void)transform:(double)aspect vect:(double*)v
{
    double glm[16];
    
    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[0*4+0] = cos(anglez.v*3.14159/180.0);
    glm[0*4+1] = -sin(anglez.v*3.14159/180.0);
    glm[1*4+0] = sin(anglez.v*3.14159/180.0);
    glm[1*4+1] = cos(anglez.v*3.14159/180.0);
    mul4(glm, v);

    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[0*4+0] = cos(angley.v*3.14159/180.0);
    glm[0*4+2] = -sin(angley.v*3.14159/180.0);
    glm[2*4+0] = sin(angley.v*3.14159/180.0);
    glm[2*4+2] = cos(angley.v*3.14159/180.0);
    mul4(glm, v);
    
    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[1*4+1] = cos(anglex.v*3.14159/180.0);
    glm[1*4+2] = -sin(anglex.v*3.14159/180.0);
    glm[2*4+1] = sin(anglex.v*3.14159/180.0);
    glm[2*4+2] = cos(anglex.v*3.14159/180.0);
    mul4(glm, v);
    
    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[12] = vec[0].v; glm[13] = vec[1].v; glm[14] = vec[2].v;
    mul4(glm, v);
    
    
    memset(&glm, 0, sizeof(glm));
    glm[0] = 2.0/(exp(zoom.v)*aspect);
    glm[5] = 2.0/exp(zoom.v);
    glm[10] = 101.0/(-99.0);
    glm[14] = 200.0/(-99.0);
    glm[11] = -1.0;
    mul4(glm, v);
    
    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[13] = offset;
    mul4(glm, v);

}

- (void)glinittransform:(double)aspect
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity(); double glm[16];
    
    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[12] = vec[0].v; glm[13] = vec[1].v; glm[14] = vec[2].v;
    glMultMatrixd(glm);
    
    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[1*4+1] = cos(anglex.v*3.14159/180.0);
    glm[1*4+2] = -sin(anglex.v*3.14159/180.0);
    glm[2*4+1] = sin(anglex.v*3.14159/180.0);
    glm[2*4+2] = cos(anglex.v*3.14159/180.0);
    glMultMatrixd(glm);
    
    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[0*4+0] = cos(angley.v*3.14159/180.0);
    glm[0*4+2] = -sin(angley.v*3.14159/180.0);
    glm[2*4+0] = sin(angley.v*3.14159/180.0);
    glm[2*4+2] = cos(angley.v*3.14159/180.0);
    glMultMatrixd(glm);

    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[0*4+0] = cos(anglez.v*3.14159/180.0);
    glm[0*4+1] = -sin(anglez.v*3.14159/180.0);
    glm[1*4+0] = sin(anglez.v*3.14159/180.0);
    glm[1*4+1] = cos(anglez.v*3.14159/180.0);
    glMultMatrixd(glm);
    
    glMatrixMode(GL_PROJECTION);

    memset(&glm, 0, sizeof(glm)); for(int i=0;i<4;i++)glm[5*i]=1.0;
    glm[13] = offset;
    glLoadMatrixd(glm);
    
    memset(&glm, 0, sizeof(glm));
    
    glm[0] = 2.0/(exp(zoom.v)*aspect);
    glm[5] = 2.0/(exp(zoom.v));
    glm[10] = 101.0/(-99.0);
    glm[14] = 200.0/(-99.0);
    glm[11] = -1.0;
    glMultMatrixd(glm);
}


-(void)loadlseries:(char*)filename width:(int)w height:(int)h
{
    glBindTexture(GL_TEXTURE_2D, lseriestexture);

    if(lseries)return;
    
    FILE * file = fopen(filename, "r");
    if(!file)return;
    
    lseries = malloc(w*h*2*sizeof(GLfloat));
    wlseries = w;
    hlseries = h;

    for(int j=0;j<h;j++)
	{
		for(int i=0;i<w;i++)
		{
            float x, y;
			fscanf(file, "%e %e", &x, &y);
            lseries[(j*w+i)*2] = x;
            lseries[(j*w+i)*2+1] = y;
		}
	}

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RG, w, h, 0, GL_RG, GL_FLOAT, lseries);
    
    fclose(file);
}

-(void)loadvideo:(char*)filename frame:(int)iframe shading:(bool)simpleshading
{
    
#define R(x) fread(&x, 1, sizeof(x), file);
    
    int videoindex[4096];
    
    FILE * file = fopen(filename,"rb");
    if(!file)return;

    glEnable(GL_LIGHTING) ;
    glEnable(GL_LIGHT0) ;
    glEnable(GL_COLOR_MATERIAL) ;
    glEnable(GL_DEPTH_TEST);
    
    if(simpleshading)glEnable(GL_CULL_FACE);else glDisable(GL_CULL_FACE);
    
    fread(&videoindex, 1, sizeof(videoindex), file);
    fseek(file, videoindex[iframe], SEEK_SET);
    
    int s; R(s); int i = 0;
	if(s)
	{
		glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
		glBegin(GL_TRIANGLES);
		while(s)
		{
            i++;
			double r, g, b;
			R(r); R(g); R(b);
			glColor3d(r, g, b);
            
			double v0[3], v1[3], v2[3], n0[3], n1[3], n2[3];
            
            R(v0);R(v1);R(v2);R(n0);R(n1);R(n2);
            
			if(simpleshading)
			{
				double n[3];
                n[0] = (v1[1]-v0[1])*(v2[2]-v0[2])-(v1[2]-v0[2])*(v2[1]-v0[1]);
                n[1] = (v1[2]-v0[2])*(v2[0]-v0[0])-(v1[0]-v0[0])*(v2[2]-v0[2]);
                n[2] = (v1[0]-v0[0])*(v2[1]-v0[1])-(v1[1]-v0[1])*(v2[0]-v0[0]);
                
                double l = sqrt(n[0]*n[0]+n[1]*n[1]+n[2]*n[2]);
                
				n[0] /= l;
				n[1] /= l;
				n[2] /= l;
                
				glNormal3dv(n);
				glVertex3dv(v0);
				glNormal3dv(n);
				glVertex3dv(v1);
				glNormal3dv(n);
				glVertex3dv(v2);
                
                n[0] *= -1;
                n[1] *= -1;
                n[2] *= -1;
                
				glNormal3dv(n);
				glVertex3dv(v0);
				glNormal3dv(n);
				glVertex3dv(v2);
				glNormal3dv(n);
				glVertex3dv(v1);
			}
			else
			{
				glNormal3dv(n0);
				glVertex3dv(v0);
				glNormal3dv(n1);
				glVertex3dv(v1);
				glNormal3dv(n2);
				glVertex3dv(v2);
			}
            
			R(s);
		}
		glEnd();
	}
    
    
	R(s);
	if(s)
	{
		glColorMaterial(GL_FRONT, GL_AMBIENT) ; //GL_SPECULAR
		glLineWidth(LINEWIDTH);
		glBegin(GL_LINES); // right
		while(s)
		{
			double r, g, b;
			double v0[3], v1[3];
            
            glNormal3d(0.0,0.0,0.0);
			R(r); R(g); R(b); glColor3d(r, g, b);
			R(v0); R(v1); glVertex3dv(v0); glVertex3dv(v1);
            
			R(s);
		}
		glEnd();
	}
    
	R(s);
	if(s)
	{
		glColorMaterial(GL_FRONT, GL_AMBIENT) ; //GL_SPECULAR
		glPointSize(POINTSIZE);
		glBegin(GL_POINTS); // right
		while(s)
		{
			double r, g, b;
			double v0[3];
            
            glNormal3d(0.0,0.0,0.0);
			R(r); R(g); R(b); glColor3d(r, g, b);
			R(v0); glVertex3dv(v0);
            
			R(s);
		}
		glEnd();
	}
    
    fclose(file);
    
    glDisable(GL_LIGHTING) ;
    glDisable(GL_LIGHT0) ;
    glDisable(GL_COLOR_MATERIAL) ;
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
}

- (void)outputtextgl:(CGRect)rect string:(NSString*)s atpos:(double*)p color:(CGFloat*)col
{
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();

    CTFontRef font = CTFontCreateWithName(CFSTR("ArialMT"), FONTSIZE, nil);
    CGColorRef color = CGColorCreate(colorSpace, col);
    
    NSDictionary* attributes = [NSDictionary dictionaryWithObjectsAndKeys:
                                (__bridge id)font, kCTFontAttributeName,
                                color, kCTForegroundColorAttributeName,
                                nil];
    
    
    NSAttributedString* as = [[NSAttributedString alloc] initWithString:s attributes:attributes];
    
    CTLineRef line = CTLineCreateWithAttributedString((CFAttributedStringRef)as);

    CGFloat ascent, descent, leading;
    double fWidth = CTLineGetTypographicBounds(line, &ascent, &descent, &leading);
    
    int windowwidth = fWidth+20;
    int windowheight = FONTHEIGHT;
    
    void* window;
    
    window = malloc(windowwidth*windowheight*4);
    memset(window, 0, windowwidth*windowheight*4);

    NSUInteger bytesPerPixel = 4;
    NSUInteger bytesPerRow = bytesPerPixel * windowwidth;
    NSUInteger bitsPerComponent = 8;
    CGContextRef cxt = CGBitmapContextCreate(window, windowwidth, windowheight,
                                             bitsPerComponent, bytesPerRow, colorSpace,
                                             kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big);
    
    CGContextSetTextPosition(cxt, 10.0, 10.0); CTLineDraw(line, cxt);
    CFRelease(line);
    
    CGColorSpaceRelease(colorSpace);
    CGContextRelease(cxt);
    CFRelease(font);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    glBindTexture(GL_TEXTURE_2D, windowtexture);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, windowwidth, windowheight, 0, GL_RGBA, GL_UNSIGNED_BYTE, window);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    double v[4] = {p[0], p[1], p[2], 1.0};
    
    [self transform:(double)rect.size.width/(double)rect.size.height vect:v];

    double dw =  2.0*(double)windowwidth / rect.size.width;
    double dh =  2.0*(double)windowheight / rect.size.height;
    
    glBegin(GL_POLYGON);
    glColor4d(1.0,1.0,1.0,1.0);
    
    glTexCoord2d(0.0,1.0);
    glVertex2d(v[0]/v[3], v[1]/v[3]);
    
    glTexCoord2d(1.0,1.0);
    glVertex2d(v[0]/v[3]+dw, v[1]/v[3]);
    
    glTexCoord2d(1.0,0.0);
    glVertex2d(v[0]/v[3]+dw, v[1]/v[3]+dh);
    
    glTexCoord2d(0.0,0.0);
    glVertex2d(v[0]/v[3], v[1]/v[3]+dh);
    glEnd();
    
    glBlendFunc(GL_ONE, GL_ZERO);

    glBindTexture(GL_TEXTURE_2D, 0);

    free(window);
}

-(void)mouseMoved:(NSEvent *)theEvent
{
    NSPoint p = [self convertPoint:[theEvent locationInWindow] fromView:nil];
    p = [self convertPointToBacking:p];

    
    p.x -= mouserect.origin.x+mouserect.size.width/2;
    p.y -= mouserect.origin.y+mouserect.size.height/2;
    
    p.x /= mouserect.size.height;
    p.y /= mouserect.size.height;
    
    p.x *= -exp(zoom.v)*vec[2].v;
    p.y *= -exp(zoom.v)*vec[2].v;
    
    p.x-=vec[0].v;
    p.y-=vec[1].v;
    
    valuegoal(&basisx, 0.0, p.x);
    valuegoal(&basisy, 0.0, p.y);
}


- (void)draw:(double)time
{
    CGLLockContext([[self openGLContext] CGLContextObj]);
    [[self openGLContext] makeCurrentContext];

    NSRect backingBounds = [self convertRectToBacking:[self bounds]];
    
    GLdouble backingPixelWidth  = (GLdouble)(backingBounds.size.width),
    backingPixelHeight = (GLdouble)(backingBounds.size.height);

    glViewport(0, 0, backingPixelWidth, backingPixelHeight);
    glScissor(0, 0, backingPixelWidth, backingPixelHeight);

    glClearColor(0, 0, 0, 255);
    glClear(GL_COLOR_BUFFER_BIT);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    condensevalue(&anglex, time);
    condensevalue(&angley, time);
    condensevalue(&anglez, time);
    condensevalue(&zoom, time);
    condensevalue(&vec[0], time);
    condensevalue(&vec[1], time);
    condensevalue(&vec[2], time);
    
    glBindTexture(GL_TEXTURE_2D, pdftexture);
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    
    glBegin(GL_POLYGON);
    glColor4d(1.0,1.0,1.0,1.0);
    glTexCoord2d(0.0,1.0);
    glVertex2d(-1.0, -1.0);
    glTexCoord2d(1.0,1.0);
    glVertex2d(1.0, -1.0);
    glTexCoord2d(1.0,0.0);
    glVertex2d(1.0, 1.0);
    glTexCoord2d(0.0,0.0);
    glVertex2d(-1.0, 1.0);
    glEnd();
    
    MBPage* mbpage = [_pages objectForKey:[NSNumber numberWithInt:ipage]];
    
    if(mbpage)
    {
        for(MBCmd *item in [mbpage commands])
        {
            if([[item cmd] isEqualToString:@"HIDE"])
            {
                for(NSString *str in [item params])
                {
                    MBBox * box = [_boxes objectForKey:str];
                    if(box)
                    {
                        CGRect rect = [box rect]; rect.origin.x -= 4.0;rect.origin.y -= 2.0; rect.size.height += 2.0;
                        rect = CGRectApplyAffineTransform(rect, transform);
                        
                        glViewport(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        glScissor(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        
                        glMatrixMode(GL_MODELVIEW);
                        glLoadIdentity();
                        
                        glMatrixMode(GL_PROJECTION);
                        glLoadIdentity();
                        
                        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                        glBindTexture(GL_TEXTURE_2D, 0);
                        
                        glBegin(GL_POLYGON);
                        glColor4d(1.0, 1.0, 1.0, 0.8);
                        glVertex2d(-1.0, -1.0);
                        glVertex2d(-1.0, 1.0);
                        glVertex2d(1.0, 1.0);
                        glVertex2d(1.0, -1.0);
                        glEnd();

                        glBlendFunc(GL_ONE, GL_ZERO);
                    }
                }
            }
            
            if([[item cmd] isEqualToString:@"OFFSET"])
            {
                offset = [[[item params] objectAtIndex:0] doubleValue];
            }
            
            if([[item cmd] isEqualToString:@"GOAL"]&&!goalscompleted)
            {
                double start = [[[item params] objectAtIndex:0] doubleValue];
                double stop = [[[item params] objectAtIndex:1] doubleValue];
   
                if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                {
                    valuegoal(&vec[0], stop==0.0?0.0:stop+frametime, [[[item params] objectAtIndex:2] doubleValue]);
                    valuegoal(&vec[1], stop==0.0?0.0:stop+frametime, [[[item params] objectAtIndex:3] doubleValue]);
                    valuegoal(&vec[2], stop==0.0?0.0:stop+frametime, [[[item params] objectAtIndex:4] doubleValue]);

                    valuegoal(&anglex, stop==0.0?0.0:stop+frametime, [[[item params] objectAtIndex:5] doubleValue]);
                    valuegoal(&angley, stop==0.0?0.0:stop+frametime, [[[item params] objectAtIndex:6] doubleValue]);
                    valuegoal(&anglez, stop==0.0?0.0:stop+frametime, [[[item params] objectAtIndex:7] doubleValue]);
                    
                    valuegoal(&zoom, stop==0.0?0.0:stop+frametime, log([[[item params] objectAtIndex:8] doubleValue]));
                    valuegoal(&basisx, stop==0.0?0.0:stop+frametime, [[[item params] objectAtIndex:9] doubleValue]);
                    valuegoal(&basisy, stop==0.0?0.0:stop+frametime, [[[item params] objectAtIndex:10] doubleValue]);
                    
                    if(stop==0.0)goalscompleted = true;
                }
            }
            
            if([[item cmd] isEqualToString:@"GLVIDEO"])
            {
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    int startframe = [[[item params] objectAtIndex:4] intValue];
                    int stopframe = [[[item params] objectAtIndex:5] intValue];
                    int iframe;
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {
                        bool simpleshading = true;
                        if ([[item params] count]>6)if(![[[item params] objectAtIndex:6] intValue])simpleshading = false;
                        
                        if(stop==0.0)
                            iframe = startframe;
                        else
                            iframe = (time-frametime)/(stop-start)*(stopframe-startframe)+startframe;
                        
                        if(iframe>stopframe)iframe = stopframe;
                        
                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        glViewport(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        glScissor(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];
                        
                        glBindTexture(GL_TEXTURE_2D, 0);
                        
                        NSString * s2 = [NSString stringWithFormat:@"%@/%@", [[_url path] stringByDeletingLastPathComponent], [[item params] objectAtIndex:3]];
                        char sc[2048];
                        [s2 getCString:sc maxLength:2048 encoding:NSASCIIStringEncoding];
                        
                        glClearColor(1.0f, 1.0f, 1.0f, 1.0f) ; glClearDepth(1.0);
                        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); //
                        
                        [self loadvideo:sc frame:iframe shading:simpleshading];
                    }
                }
            }
            
            if([[item cmd] isEqualToString:@"POINT"])
            {
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {
                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];

                        glColorMaterial(GL_FRONT, GL_AMBIENT) ; //GL_SPECULAR
                        glPointSize(POINTSIZE);
                        glBegin(GL_POINTS); // right
                        glNormal3d(0.0,0.0,0.0);
                        
                        GLubyte r = [[[item params] objectAtIndex:6] intValue];
                        GLubyte g = [[[item params] objectAtIndex:7] intValue];
                        GLubyte b = [[[item params] objectAtIndex:8] intValue];
                        CGFloat c[4] = {r/256.0,g/256.0,b/256.0,1.0};
                        
                        GLdouble v[3];
                        v[0] = [[[item params] objectAtIndex:3] doubleValue];
                        v[1] = [[[item params] objectAtIndex:4] doubleValue];
                        v[2] = [[[item params] objectAtIndex:5] doubleValue];
                        
                        glColor3ub(r,g,b);
                        glVertex3dv(v);
                        
                        glEnd();
                        
                        if([[item params] count]>=10)
                            [self outputtextgl:rect string:[[item params] objectAtIndex:9] atpos:v color:c];
                        
                    }
                }
            }
            
            if([[item cmd] isEqualToString:@"LINE"])
            {
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {
                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];
                        
                        glColorMaterial(GL_FRONT, GL_AMBIENT) ; //GL_SPECULAR
                        glLineWidth(LINEWIDTH);
                        glBegin(GL_LINES); // right
                        glNormal3d(0.0,0.0,0.0);
                        
                        GLubyte r = [[[item params] objectAtIndex:9] intValue];
                        GLubyte g = [[[item params] objectAtIndex:10] intValue];
                        GLubyte b = [[[item params] objectAtIndex:11] intValue];
                        
                        GLdouble v0 = [[[item params] objectAtIndex:3] doubleValue];
                        GLdouble v1 = [[[item params] objectAtIndex:4] doubleValue];
                        GLdouble v2 = [[[item params] objectAtIndex:5] doubleValue];
                        
                        GLdouble w0 = [[[item params] objectAtIndex:6] doubleValue];
                        GLdouble w1 = [[[item params] objectAtIndex:7] doubleValue];
                        GLdouble w2 = [[[item params] objectAtIndex:8] doubleValue];
                        
                        GLdouble f = 1000.0;
                        
                        GLdouble z0 = w0-v0;
                        GLdouble z1 = w1-v1;
                        GLdouble z2 = w2-v2;
                        
                        glColor3ub(r,g,b);
                        glVertex3d(v0-z0*f,v1-z1*f,v2-z2*f);
                        glVertex3d(w0+z0*f,w1+z1*f,w2+z2*f);
                        
                        glEnd();
                        
                    }
                }
            }
            
            if([[item cmd] isEqualToString:@"TESTCIRCLE"])
            {
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {
                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        mouserect = rect;
                        
                        glViewport(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        glScissor(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];
                        
                        glColorMaterial(GL_FRONT, GL_AMBIENT) ; //GL_SPECULAR
                        glLineWidth(LINEWIDTH);
                        glBegin(GL_LINES); // right
                        glNormal3d(0.0,0.0,0.0);
                        glColor3d(1.0,1.0,1.0);
                        
                        for(int i=0;i<360;i++)
                        {
                            glVertex2d(basisx.v+sin(i*M_PI/180.0), basisy.v+cos(i*M_PI/180.0));
                            glVertex2d(basisx.v+sin((i+1)*M_PI/180.0), basisy.v+cos((i+1)*M_PI/180.0));
                        }
                        
                        glEnd();
                        
                    }
                }
            }
            if([[item cmd] isEqualToString:@"TESTCIRCLEIMAGE"]&&lseries)
            {
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {
                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        glViewport(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        glScissor(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];
                        
                        glColorMaterial(GL_FRONT, GL_AMBIENT);
                        glLineWidth(LINEWIDTH);
                        glBegin(GL_LINES);
                        glNormal3d(0.0,0.0,0.0);
                        glColor3d(1.0,1.0,1.0);
                        
                        double d = [[[item params] objectAtIndex:3] doubleValue];
                        
                        for(int k=0;k<360;k++)
                        {
                            float x = basisx.v+sin(k*M_PI/180.0);
                            float y = basisy.v+cos(k*M_PI/180.0);
                            
                            int i = ((x+6.0) / 12.0)*wlseries;
                            int j = ((y+6.0) / 12.0)*hlseries;
                            
                            if(i>=0&&i<wlseries&&j>=0&&j<hlseries)
                            {
                                x = basisx.v+sin((k+1)*M_PI/180.0);
                                y = basisy.v+cos((k+1)*M_PI/180.0);
                                
                                int i2 = ((x+6.0) / 12.0)*wlseries;
                                int j2 = ((y+6.0) / 12.0)*hlseries;
                                
                                if(i2>=0&&i2<wlseries&&j2>=0&&j2<hlseries)
                                {
                                    glVertex2d(lseries[(j*wlseries+i)*2]*d, lseries[(j*wlseries+i)*2+1]*d);
                                    glVertex2d(lseries[(j2*wlseries+i2)*2]*d, lseries[(j2*wlseries+i2)*2+1]*d);
                                }
                            }
                            
                        }
                        
                        glEnd();
                        
                    }
                }
            }
            if([[item cmd] isEqualToString:@"ARROW"])
            {
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {
                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];
                        
                        glColorMaterial(GL_FRONT, GL_AMBIENT) ; //GL_SPECULAR
                        glLineWidth(LINEWIDTH);
                        glBegin(GL_LINES); // right
                        glNormal3d(0.0,0.0,0.0);
                        
                        GLubyte r = [[[item params] objectAtIndex:12] intValue];
                        GLubyte g = [[[item params] objectAtIndex:13] intValue];
                        GLubyte b = [[[item params] objectAtIndex:14] intValue];
                        
                        GLdouble v0 = [[[item params] objectAtIndex:3] doubleValue];
                        GLdouble v1 = [[[item params] objectAtIndex:4] doubleValue];
                        GLdouble v2 = [[[item params] objectAtIndex:5] doubleValue];
                        
                        GLdouble w0 = [[[item params] objectAtIndex:6] doubleValue];
                        GLdouble w1 = [[[item params] objectAtIndex:7] doubleValue];
                        GLdouble w2 = [[[item params] objectAtIndex:8] doubleValue];
                        
                        GLdouble n0 = [[[item params] objectAtIndex:9] doubleValue];
                        GLdouble n1 = [[[item params] objectAtIndex:10] doubleValue];
                        GLdouble n2 = [[[item params] objectAtIndex:11] doubleValue];


                        GLdouble z0 = w0-v0;
                        GLdouble z1 = w1-v1;
                        GLdouble z2 = w2-v2;
                        
                        GLdouble zl = sqrt(z0*z0+z1*z1+z2*z2);
                        GLdouble nl = sqrt(n0*n0+n1*n1+n2*n2);
                        
                        z0/=zl; z1/=zl; z2/=zl;

                        
                        glColor3ub(r,g,b);
                        glVertex3d(v0,v1,v2);
                        glVertex3d(w0,w1,w2);

                        glVertex3d(w0-z0*nl+(z1*n2-z2*n1),w1-z1*nl+(z2*n0-z0*n2),w2-z2*nl+(z0*n1-z1*n0));
                        glVertex3d(w0,w1,w2);

                        glVertex3d(w0-z0*nl-(z1*n2-z2*n1),w1-z1*nl-(z2*n0-z0*n2),w2-z2*nl-(z0*n1-z1*n0));
                        glVertex3d(w0,w1,w2);

                        glEnd();
                        
                    }
                }
            }
            
            if([[item cmd] isEqualToString:@"LSERIES"]||[[item cmd] isEqualToString:@"LSERIESC"])
            {
                
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {
                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        glViewport(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        glScissor(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];
                        
                        glUseProgramObjectARB(proglseries);
                        
                        NSString * s2 = [NSString stringWithFormat:@"%@/%@", [[_url path] stringByDeletingLastPathComponent], [[item params] objectAtIndex:3]];
                        char sc[2048];
                        [s2 getCString:sc maxLength:2048 encoding:NSASCIIStringEncoding];
                        
                        glClearColor(1.0f, 1.0f, 1.0f, 1.0f) ; glClearDepth(1.0);
                        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); //
                        
                        [self loadlseries:sc width:[[[item params] objectAtIndex:4] intValue] height:[[[item params] objectAtIndex:5] intValue]];
                        
                        GLint loc;
                        
                        loc = glGetUniformLocationARB(proglseries, "lseries");
                        if(loc != -1) glUniform1i(loc, 0);
                        loc = glGetUniformLocationARB(proglseries, "mode");
                        if(loc != -1) glUniform1i(loc, [[item cmd] isEqualToString:@"LSERIES"]?0:1);
                        
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
                        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
                        
                        glClearColor(255, 255, 255, 255);
                        glClear(GL_COLOR_BUFFER_BIT);
                        
                        float d = 6.0;
                        
                        glBegin(GL_POLYGON);
                        
                        glTexCoord2d(0.0, 0.0);
                        glVertex2d(-d, -d);
                        
                        glTexCoord2d(0.0, 1.0);
                        glVertex2d(-d, +d);
                        
                        glTexCoord2d(1.0, 1.0);
                        glVertex2d(+d, +d);
                        
                        glTexCoord2d(1.0, 0.0);
                        glVertex2d(+d, -d);
                        
                        glEnd();
                        
                        glBindTexture(GL_TEXTURE_2D, 0);
                        
                        glUseProgramObjectARB(0);
                        
                    }
                }
            }
            
            if([[item cmd] isEqualToString:@"IDENTITY"])
            {
                
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {
                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        glViewport(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        glScissor(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];
                        
                        glUseProgramObjectARB(proglseries);
                        
                        glClearColor(1.0f, 1.0f, 1.0f, 1.0f) ; glClearDepth(1.0);
                        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); //
                        
                        GLint loc;
                        
                        loc = glGetUniformLocationARB(proglseries, "lseries");
                        if(loc != -1) glUniform1i(loc, 0);
                        loc = glGetUniformLocationARB(proglseries, "mode");
                        if(loc != -1) glUniform1i(loc, 2);
                        
                        glClearColor(255, 255, 255, 255);
                        glClear(GL_COLOR_BUFFER_BIT);
                        
                        float d = 6.0;
                        
                        glBegin(GL_POLYGON);
                        
                        glTexCoord2d(0.0, 0.0);
                        glVertex2d(-d, -d);
                        
                        glTexCoord2d(0.0, 1.0);
                        glVertex2d(-d, +d);
                        
                        glTexCoord2d(1.0, 1.0);
                        glVertex2d(+d, +d);
                        
                        glTexCoord2d(1.0, 0.0);
                        glVertex2d(+d, -d);
                        
                        glEnd();
                        
                        glBindTexture(GL_TEXTURE_2D, 0);
                        
                        glUseProgramObjectARB(0);
                        
                    }
                }
            }

            if([[item cmd] isEqualToString:@"HYPERBOLIC"])
            {
                MBBox * box = [_boxes objectForKey:[[item params] objectAtIndex:2]];
                if(box)
                {
                    
                    double start = [[[item params] objectAtIndex:0] doubleValue];
                    double stop = [[[item params] objectAtIndex:1] doubleValue];
                    
                    if(time-frametime>=start && (stop==0.0||time-frametime<stop))
                    {

                        CGRect rect = CGRectApplyAffineTransform([box rect], transform);
                        
                        mouserect = rect;
                        
                        glViewport(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        glScissor(rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
                        
                        [self glinittransform:(double)rect.size.width/(double)rect.size.height];
                        
                        glUseProgramObjectARB(proghyperbolic);
                        
                        GLint loc;
                        
                        loc = glGetUniformLocationARB(proghyperbolic, "linewidth");
                        if(loc != -1) glUniform1f(loc, GLLINEWIDTH/exp(zoom.v));
                        loc = glGetUniformLocationARB(proghyperbolic, "basisx");
                        if(loc != -1) glUniform1f(loc, basisx.v);
                        loc = glGetUniformLocationARB(proghyperbolic, "basisy");
                        if(loc != -1) glUniform1f(loc, basisy.v);
                        loc = glGetUniformLocationARB(proghyperbolic, "hyperbolicmode");
                        if(loc != -1) glUniform1i(loc,[[[item params] objectAtIndex:3] intValue]);
                        loc = glGetUniformLocationARB(proghyperbolic, "lattice");
                        if(loc != -1) glUniform1i(loc, [[[item params] objectAtIndex:4] intValue]);
                        loc = glGetUniformLocationARB(proghyperbolic, "fundamentalmesh");
                        if(loc != -1) glUniform1i(loc, [[[item params] objectAtIndex:5] intValue]);
                        
                        
                        glClearColor(255, 255, 255, 255);
                        glClear(GL_COLOR_BUFFER_BIT);
                        
                        glBegin(GL_POLYGON);
                        
                        glTexCoord2d(-10.0, -10.0);
                        glVertex2d(-10.0, -10.0);
                        
                        glTexCoord2d(-10.0, 10.0);
                        glVertex2d(-10.0, 10.0);
                        
                        glTexCoord2d(10.0, 10.0);
                        glVertex2d(10.0, 10.0);
                        
                        glTexCoord2d(10.0, -10.0);
                        glVertex2d(10.0, -10.0);
                        
                        glEnd();
                        
                        glUseProgramObjectARB(0);
                    }
                }
            }
        }
    }

    glFlush();

    [[self openGLContext] flushBuffer];
    CGLUnlockContext([[self openGLContext] CGLContextObj]);
}

- (BOOL)acceptsFirstResponder
{
    return YES;
}

- (void)settitle
{
    [[self window] setTitle:[NSString stringWithFormat:@"%@ (%i/%i)", pdftitle, ipage, npage]];
}

void ApplierFunction(const char *key, CGPDFObjectRef value, void *info)
{
    MainView *mv = (__bridge MainView*) info;
    
    CGPDFStringRef pdfstr;
    if( CGPDFObjectGetValue( value, kCGPDFObjectTypeString, &pdfstr ) )
    {
        if(strstr(key, "Title")==key)
        {
            CFStringRef _res = CGPDFStringCopyTextString(pdfstr);
            mv->pdftitle = [NSString stringWithString:(__bridge NSString *)_res];
            CFRelease(_res);
        }
        
        if(strstr(key, "MB_CMD")==key)
        {
            char string[1024], name[1024]; strncpy(string, (const char*)CGPDFStringGetBytePtr( pdfstr ), sizeof(string));
            char delimiter[] = " ";
            char* ptr;
            ptr = strtok(string, delimiter);
            
            if(ptr!=NULL && !strcmp(ptr, "BOX"))
            {
                ptr = strtok(NULL, "");
                if(ptr)
                {
                    int i[5];
                    if(sscanf(ptr, "%s %i %i %i %i %i", (char*)name, &i[0], &i[1], &i[2], &i[3], &i[4])==6)
                    {
                        CGRect rect = CGRectMake(i[1]/65536.0, i[2]/65536.0, i[3]/65536.0, i[4]/65536.0);
                        
                        MBBox *box = [MBBox alloc];
                        [box setIpdfpage:i[0]];
                        [box setRect:rect];
                        
                        [[mv boxes] setObject:box forKey:[NSString stringWithUTF8String:name]];
                    }
                }
            }
            else
            if(ptr!=NULL && !strcmp(ptr, "PAGE"))
            {
                ptr = strtok(NULL, "");
                if(ptr)
                {
                    int ipage, ipdfpage;
                    if(sscanf(ptr, "%i %i", &ipage, &ipdfpage)==2)
                    {
                        MBPage *page = [MBPage alloc];
                        [page setIpdfpage:ipdfpage];
                        [page setCommands:[[NSMutableArray alloc] init]];
                        
                        [[mv pages] setObject:page forKey:[NSNumber numberWithInt:ipage]];
                        mv->npage ++;
                    }
                }
            }
            else
            if(ptr!=NULL)
            {
                MBCmd *cmd = [MBCmd alloc]; MBPage *page;
                [cmd setCmd:[NSString stringWithUTF8String:ptr]];
                [cmd setParams:[[NSMutableArray alloc] init]];
                
                int iparam = 0;
                
                ptr = strtok(NULL, delimiter);
                while(ptr)
                {
                    if(iparam==0)
                    {
                        int i;
                        if(sscanf(ptr, "%i", &i)==1)
                        {
                            page = [[mv pages] objectForKey:[NSNumber numberWithInt:i]];
                        }
                    }
                    else
                        [[cmd params] addObject:[NSString stringWithUTF8String:ptr]];
                    ptr = strtok(NULL, delimiter); iparam++;
                }
                if(page)
                 {
                     [[page commands] addObject:cmd];
                 }
            }
        }
    }
}

- (void)pdfreadmetadata
{
    CGPDFDocumentRef pdfDocument = CGPDFDocumentCreateWithURL((__bridge CFURLRef)[self url]);
    CGPDFDictionaryRef dict = CGPDFDocumentGetInfo(pdfDocument);
    
    npage = 0;
    CGPDFDictionaryApplyFunction(dict, &ApplierFunction, (__bridge void *) self);
    CGPDFDocumentRelease(pdfDocument);
    
    [self settitle];
}

- (void)pdfloadpage
{
    if(lseries){free(lseries); lseries = NULL;}
    
    NSRect backingBounds = [self convertRectToBacking:[self bounds]];
    
    GLdouble backingPixelWidth  = (GLdouble)(backingBounds.size.width),
    backingPixelHeight = (GLdouble)(backingBounds.size.height);
    
    MBPage* mbpage = [_pages objectForKey:[NSNumber numberWithInt:ipage]];
    
    if(!mbpage)
    {
        CGLLockContext([[self openGLContext] CGLContextObj]);
        [[self openGLContext] makeCurrentContext];
        
        glBindTexture(GL_TEXTURE_2D, pdftexture);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, backingPixelWidth, backingPixelHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
        
        CGLUnlockContext([[self openGLContext] CGLContextObj]);

        return;
    }
    
    CGPDFDocumentRef pdfDocument = CGPDFDocumentCreateWithURL((__bridge CFURLRef)[self url]);
    CGPDFPageRef page = CGPDFDocumentGetPage(pdfDocument, [mbpage ipdfpage]);
    
    CGRect cropBox = CGPDFPageGetBoxRect(page, kCGPDFCropBox);
    
    NSInteger rotationAngle = CGPDFPageGetRotationAngle(page);
    CGFloat angleInRadians = -rotationAngle * (M_PI / 180);
    transform = CGAffineTransformMakeRotation(angleInRadians);
    CGRect rotatedCropRect = CGRectApplyAffineTransform(cropBox, transform);
    
    CGFloat scale = MIN(backingPixelHeight / CGRectGetHeight(rotatedCropRect), backingPixelWidth / CGRectGetWidth(rotatedCropRect));
    
    size_t bytesPerRow = ((size_t)backingPixelWidth)*4;
    
    void *data = malloc(bytesPerRow*backingPixelHeight*4);
    
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    CGContextRef context =  CGBitmapContextCreate(data, backingPixelWidth, backingPixelHeight, 8, bytesPerRow, colorSpace, (kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big));
    CGColorSpaceRelease(colorSpace);
    
    CGContextSetRGBFillColor(context, 0.0, 0.0, 0.0, 1.0);
    CGContextFillRect(context, CGRectMake(0.0, 0.0, backingPixelWidth, backingPixelHeight));
    
    transform = CGPDFPageGetDrawingTransform(page, kCGPDFCropBox, CGRectMake(0, 0, backingPixelWidth, backingPixelHeight), 0, true);
    
    if (scale > 1)
    {
        transform = CGAffineTransformTranslate(transform, CGRectGetMidX(cropBox), CGRectGetMidY(cropBox));
        transform = CGAffineTransformScale(transform, scale, scale);
        transform = CGAffineTransformTranslate(transform, -CGRectGetMidX(cropBox), -CGRectGetMidY(cropBox));
    }
    
    CGContextConcatCTM(context, transform);
    
    CGContextAddRect(context, cropBox);
    CGContextClip(context);
    
    CGContextDrawPDFPage(context, page);
    
    CGContextRelease(context);
    CGPDFDocumentRelease(pdfDocument);
    
    CGLLockContext([[self openGLContext] CGLContextObj]);
    [[self openGLContext] makeCurrentContext];
    
    glBindTexture(GL_TEXTURE_2D, pdftexture);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, backingPixelWidth, backingPixelHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

    CGLUnlockContext([[self openGLContext] CGLContextObj]);

    free(data);
}

- (void)viewDidMoveToWindow
{
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(windowResized:) name:NSWindowDidResizeNotification object:[self window]];
}

- (void)mouseDown:(NSEvent *)theEvent
{
    if(displaymouse)
    {
        CGDisplayHideCursor(kCGDirectMainDisplay);
        displaymouse = false;
    }
    else
    {
        CGDisplayShowCursor(kCGDirectMainDisplay);
        displaymouse = true;
    }
}

double MachTimeToSecs(uint64_t time)
{
    mach_timebase_info_data_t timebase;
    mach_timebase_info(&timebase);
    return (double)time * (double)timebase.numer /
    (double)timebase.denom / 1e9;
}

static CVReturn MyDisplayLinkCallback(CVDisplayLinkRef displayLink, const CVTimeStamp* now, const CVTimeStamp* outputTime,
                                      CVOptionFlags flagsIn, CVOptionFlags* flagsOut, void* displayLinkContext)
{
    CVReturn result = [(__bridge MainView*)displayLinkContext getFrameForTime:outputTime current:now];
    return result;
}

- (void)keyUp:(NSEvent *)theEvent
{
    CGLLockContext([[self openGLContext] CGLContextObj]);

    bool handled = false;
    double time = [theEvent timestamp];
    
#define K(k,v,a) if([theEvent.characters isEqualToString:@k]){ condensevalue(&v, time); valuesetdv(&v, 0.0); handled = true; }

    if(![theEvent isARepeat])
    {
        K("j",angley,-90.0);
        K("k",angley,90.0);
        K("i",anglex,-90.0);
        K("m",anglex,90.0);
        K("v",anglez,-90.0);
        K("b",anglez,90.0);
        
        K("a",vec[0],-5.0);
        K("s",vec[0],5.0);
        K("w",vec[1],-5.0);
        K("y",vec[1],5.0);
        K("q",vec[2],-5.0);
        K("e",vec[2],5.0);
        
        K("1",zoom,-0.1);
        K("2",zoom,0.1);
    }
    else
        handled = true;

#undef K
    
    if ([theEvent modifierFlags] & NSNumericPadKeyMask)
    { // arrow keys have this mask
        
        NSString *theArrow = [theEvent charactersIgnoringModifiers];
        
        unichar keyChar = 0;
        
        if ( [theArrow length] == 1 ) {
            
            keyChar = [theArrow characterAtIndex:0];
            
            if ( keyChar == NSLeftArrowFunctionKey ) {
                if(![theEvent isARepeat])
                {
                    condensevalue(&anglez, time); valuesetdv(&anglez, 0.0); handled = true;
                } else
                handled = true;
                
            }
            
            if ( keyChar == NSRightArrowFunctionKey ) {
                if(![theEvent isARepeat])
                {
                    condensevalue(&anglez, time); valuesetdv(&anglez, 0.0); handled = true;
                }
                else handled = true;
                
            }
        }
    }

    
    if(!handled)[super keyUp:theEvent];
    
    CGLUnlockContext([[self openGLContext] CGLContextObj]);

}

- (void)keyDown:(NSEvent *)theEvent
{
    CGLLockContext([[self openGLContext] CGLContextObj]);

    bool handled = false;
    double time = [theEvent timestamp];
    
#define K(k,v,a) if([theEvent.characters isEqualToString:@k]){ condensevalue(&v, time); valuesetdv(&v, a); handled = true; }

    if(![theEvent isARepeat])
    {
        K("j",angley,-90.0);
        K("k",angley,90.0);
        K("i",anglex,-90.0);
        K("m",anglex,90.0);
        K("v",anglez,-90.0);
        K("b",anglez,90.0);
        
        K("a",vec[0],-1.0*exp(zoom.v));
        K("s",vec[0],1.0*exp(zoom.v));
        K("w",vec[1],-1.0*exp(zoom.v));
        K("y",vec[1],1.0*exp(zoom.v));
        K("q",vec[2],-1.0*exp(zoom.v));
        K("e",vec[2],1.0*exp(zoom.v));
        
        K("1",zoom,-1.0);
        K("2",zoom,1.0);
    }
    else
        handled = true;
    
#undef K

    if ([theEvent modifierFlags] & NSNumericPadKeyMask)
    { // arrow keys have this mask
        
        NSString *theArrow = [theEvent charactersIgnoringModifiers];
        
        unichar keyChar = 0;
        
        if ( [theArrow length] == 1 ) {
            
            keyChar = [theArrow characterAtIndex:0];
            
            if ( keyChar == NSLeftArrowFunctionKey ) {
                
                if(![theEvent isARepeat])
                {
                    condensevalue(&anglez, time); valuesetdv(&anglez, -90.0); handled = true;
                }
                else
                    handled = true;
                
            }
            
            if ( keyChar == NSRightArrowFunctionKey ) {
                
                if(![theEvent isARepeat])
                {
                    condensevalue(&anglez, time); valuesetdv(&anglez, 90.0); handled = true;
                }
                else
                handled = true;
                
            }
            
            if ( keyChar == NSUpArrowFunctionKey ) {
                
                ipage--; if(ipage<=0)ipage = 1;
                
                frametime = time; goalscompleted = false;
                
                [self pdfloadpage];
                [self settitle];

                
                handled = true;
                
            }
            
            if ( keyChar == NSDownArrowFunctionKey) {
                
                ipage++; if(ipage>npage)ipage = npage;
                
                frametime = time; goalscompleted = false;
                
                [self pdfloadpage];
                [self settitle];

                
                handled = true;
                
            }
            
        }
        
    }
    
    if(!handled)[super keyDown:theEvent];
    
    CGLUnlockContext([[self openGLContext] CGLContextObj]);

}

GLhandleARB setup_shader(const char *fname)
{
	FILE *fp;
	long len; GLhandleARB sdr, prog;
	char *src_buf;
	GLint success, linked;
    
	if(!(fp = fopen(fname, "r"))) {
		fprintf(stderr, "failed to open shader: %s\n", fname);
		return 0;
	}
	fseek(fp, 0, SEEK_END);
	len = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	src_buf = malloc(len + 1);
    
	fread(src_buf, 1, len, fp);
	src_buf[len] = 0;
    
	sdr = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
	glShaderSourceARB(sdr, 1, (const char**)&src_buf, 0);
	free(src_buf);
    
	glCompileShaderARB(sdr);
	glGetObjectParameterivARB(sdr, GL_OBJECT_COMPILE_STATUS_ARB, &success);
	if(!success) {
		int info_len;
		char *info_log;
		
		glGetObjectParameterivARB(sdr, GL_OBJECT_INFO_LOG_LENGTH_ARB, &info_len);
		if(info_len > 0) {
			if(!(info_log = malloc(info_len + 1))) {
				perror("malloc failed");
				return 0;
			}
			glGetInfoLogARB(sdr, info_len, 0, info_log);
			fprintf(stderr, "shader compilation failed: %s\n", info_log);
			free(info_log);
		} else {
			fprintf(stderr, "shader compilation failed\n");
		}
		return 0;
	}
    
	prog = glCreateProgramObjectARB();
	glAttachObjectARB(prog, sdr);
	glLinkProgramARB(prog);
	glGetObjectParameterivARB(prog, GL_OBJECT_LINK_STATUS_ARB, &linked);
	if(!linked) {
		fprintf(stderr, "shader linking failed\n");
		return 0;
	}
    
	return prog;
}

- (void)prepareOpenGL
{
    [super prepareOpenGL];
    
    GLint swapInt = 1;
    [[self openGLContext] setValues:&swapInt forParameter:NSOpenGLCPSwapInterval];
    
    CVDisplayLinkCreateWithActiveCGDisplays(&displayLink);
    
    CVDisplayLinkSetOutputCallback(displayLink, &MyDisplayLinkCallback, (__bridge void *)(self));
    
    CGLContextObj cglContext = [[self openGLContext] CGLContextObj];
    CGLPixelFormatObj cglPixelFormat = [[self pixelFormat] CGLPixelFormatObj];
    CVDisplayLinkSetCurrentCGDisplayFromOpenGLContext(displayLink, cglContext, cglPixelFormat);
    
    [[self openGLContext] makeCurrentContext];
    
    proghyperbolic = setup_shader([[[NSBundle mainBundle] pathForResource:@"hyperbolic" ofType:@"glsl"] UTF8String]);
    proglseries = setup_shader([[[NSBundle mainBundle] pathForResource:@"lseries" ofType:@"glsl"] UTF8String]);
    
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_SCISSOR_TEST);
    
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
    
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ZERO);
    glDisable(GL_DEPTH_TEST);
    
    glGenTextures(1, &pdftexture);
    glGenTextures(1, &windowtexture);
    glGenTextures(1, &lseriestexture);
    
    CVDisplayLinkStart(displayLink);
}

- (void)windowResized:(NSNotification *)notification;
{
    CGLLockContext([[self openGLContext] CGLContextObj]);

    [self pdfloadpage];

    CGLUnlockContext([[self openGLContext] CGLContextObj]);
}


- (CVReturn)getFrameForTime:(const CVTimeStamp*)outputTime current:(const CVTimeStamp*)currentTime
{
    double time = MachTimeToSecs(outputTime->hostTime);
    
    [self draw:time];
        
    return kCVReturnSuccess;
}


- (void)dealloc
{
    CVDisplayLinkStop(displayLink);
    CVDisplayLinkRelease(displayLink);
    
    if(lseries)free(lseries);
    
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}

@end
