#version 110

// hyperbolic.glsl
// beamermath

/*
 Created on 9.8.2018 by
 
 PD Dr. Fritz Hörmann
 Mathematisches Institut - Raum 433
 Albert-Ludwigs-Universität Freiburg
 Ernst-Zermelo-Strasse 1
 D-79104 Freiburg
 eMail: fritz.hoermann@math.uni-freiburg.de
 Tel: +49 761 99520388
 
 Licensed under CC BY-SA 4.0.
 https://creativecommons.org/licenses/by-sa/4.0/legalcode
 */


/* CHANNEL_ITER CHANNEL_X CHANNEL_Y */

#define M_PI 3.1415926535897932384626433832795

uniform float linewidth;

uniform float basisx;
uniform float basisy;

uniform int hyperbolicmode;
uniform bool lattice;
uniform bool fundamentalmesh;


vec4 hyperbolic(vec2 cc)
{
    if(lattice||fundamentalmesh)
    {
        float x = 1.0/basisy * (basisy*cc.x - basisx*cc.y) + 0.5;
        float y = 1.0/basisy * (cc.y) + 0.5;
        
        float xx = (fract(x)-0.5) + basisx*(fract(y)-0.5);
        float yy = basisy*(fract(y)-0.5);
        
        if(lattice)
        {
            if(int(floor(x))==0 && int(floor(y))==1)
            {
                if(xx*xx+yy*yy<20.0/linewidth*1.0/linewidth)return vec4(1.0,0.0,0.0,1.0);
            }
            else
            {
                if(xx*xx+yy*yy<20.0/linewidth*1.0/linewidth)return vec4(0.0,0.5,0.0,1.0);
            }
        }

        if(fundamentalmesh)
        {
            if(int(floor(x-0.5))==0 && int(floor(y-0.5))==0)return vec4(0.0,0.0,0.5,1.0);
        }
    }
    


    if(hyperbolicmode!=0)
    {
    
        vec2 dd = vec2(1.0,0.0), cs, ddn, cn, c = cc;
        
        if(c.y<=0.0){return vec4(1.0,1.0,1.0,1.0);}
        
        int _a=1,_b=0,_c=0,_d=1, iter=0;
        
        int k;
        int where = 0;
        
        /*
         ab     1   1 k
         cd  -1       1
         
         */
        
        k = int(floor(c.x+0.5));
        _b = k*_a+_b; _d = k*_c+_d;
        c.x -= float(k); if(k<0)iter -= k;else iter += k;
        
        while((c.x*c.x+c.y*c.y)<1.0&&iter<1000)
        {
            cs.x = c.x*c.x-c.y*c.y; cs.y = 2.0*c.x*c.y;
            
            ddn.x = dd.x*cs.x+dd.y*cs.y; ddn.y = -dd.x*cs.y+dd.y*cs.x;
            dd = ddn / (cs.x*cs.x+cs.y*cs.y);
            
            cn.x = -c.x; cn.y = c.y;
            c = cn / (c.x*c.x+c.y*c.y);
            
            int __a = _a, __c = _c; _a = -_b; _b = __a; _c = -_d; _d = __c;
            iter++;
            
            k = int(floor(c.x+0.5));
            _b = k*_a+_b; _d = k*_c+_d;
            c.x -= float(k); if(k<0)iter -= k;else iter += k;
        }
        
        float d1 = sqrt(c.x*c.x+c.y*c.y)-1.0, len = sqrt(dd.x*dd.x+dd.y*dd.y), d=d1;
        float d2 = abs(c.x-0.5); if(d2<d)d = d2;
        float d3 = abs(c.x+0.5); if(d3<d)d = d3;
        float dist = d/len;
        
        vec4 color = vec4(1.0,1.0,1.0,1.0);
        if((1.0-linewidth*dist)>0.0)color *= (1.0-linewidth*dist); else color = vec4(0.0,0.0,0.0,1.0);
        color = vec4(1.0,1.0,1.0,2.0)-color;
        
        if(hyperbolicmode==3)
        {
            if(_a==1 &&_b==0&&_c==0 &&_d==1 &&(d2/len*linewidth<4.0))color = vec4(0.0,0.0,1.0,1.0);
            if(_a==0 &&_b==1&&_c==-1&&_d==1 &&(d1/len*linewidth<4.0))color = vec4(0.0,0.0,1.0,1.0);
            if(_a==-1&&_b==0&&_c==-2&&_d==-1&&(d3/len*linewidth<4.0))color = vec4(0.0,0.0,1.0,1.0);
            
            if(_a==0 &&_b==1&&_c==-1&&_d==-1&&(d1/len*linewidth<4.0))color = vec4(0.0,0.0,1.0,1.0);
            if(_a==1 &&_b==0&&_c==0 &&_d==1 &&(d3/len*linewidth<4.0))color = vec4(0.0,0.0,1.0,1.0);
            if(_a==-1&&_b==0&&_c==2 &&_d==-1&&(d2/len*linewidth<4.0))color = vec4(0.0,0.0,1.0,1.0);

            if(_a==-1&&_b==0&&_c==2 &&_d==-1&&(d3/len*linewidth<4.0))color = vec4(0.0,1.0,1.0,1.0);
            if(_a==-1&&_b==0&&_c==-2&&_d==-1&&(d2/len*linewidth<4.0))color = vec4(1.0,0.0,0.0,1.0);
            
            if(_a==0&&_b==1&&_c==-1&&_d==4 &&(d3/len*linewidth<4.0))color = vec4(1.0,0.0,0.0,1.0);
            if(_a==0&&_b==1&&_c==-1&&_d==-4&&(d2/len*linewidth<4.0))color = vec4(0.0,1.0,1.0,1.0);
            
            if(_a==0&&_b==1&&_c==-1&&_d==4 &&(d1/len*linewidth<4.0))color = vec4(0.0,0.5,0.0,1.0);
            if(_a==0&&_b==1&&_c==-1&&_d==-4&&(d1/len*linewidth<4.0))color = vec4(1.0,0.0,1.0,1.0);

            if(_a==0&&_b==1&&_c==-1&&_d==3 &&(d1/len*linewidth<4.0))color = vec4(1.0,0.0,1.0,1.0);
            if(_a==0&&_b==1&&_c==-1&&_d==-3&&(d1/len*linewidth<4.0))color = vec4(0.0,0.5,0.0,1.0);

        }

        
        /*
         (1 basisx  (x) = Re z
         0 basisy) (y)   Im z
         
         1/basisy (basisy -basisx ) (Re z)
         0      1        (Im z)
         */
        
        if(hyperbolicmode==2)
        {
            color -= vec4(0.25+0.125*sin(float(iter)*3.8),0.25+0.125*cos(float(iter)),0.25+0.125*sin(2.11*float(iter)),1.0);
        }
        if(hyperbolicmode==3)
        {
            if((_a==1&&_b==0&&_c==0&&_d==1)
            ||(_a==0&&_b==1&&_c==-1&&_d==0)
               ||(_a==0&&_b==1&&_c==-1&&_d==-1)
               ||(_a==0&&_b==1&&_c==-1&&_d==-2)
               ||(_a==0&&_b==1&&_c==-1&&_d==-3)
               ||(_a==0&&_b==1&&_c==-1&&_d==-4)
               ||(_a==-1&&_b==0&&_c==2&&_d==-1)
                ||(_a==0&&_b==1&&_c==-1&&_d==1)
                ||(_a==0&&_b==1&&_c==-1&&_d==2)
               ||(_a==0&&_b==1&&_c==-1&&_d==3)
               ||(_a==0&&_b==1&&_c==-1&&_d==4)
                ||(_a==-1&&_b==0&&_c==-2&&_d==-1))
                color -= vec4(0.25,0.25,0.25,0.0);

        }
        if(hyperbolicmode==4)
        {
            if(_a==1&&_b==0&&_c==0&&_d==1)
                color -= vec4(0.25,0.25,0.25,0.0);
        }

        
        return color;
    }
    else
    {
        return vec4(1.0,1.0,1.0,1.0);
    }
 

}


void main()
{
    gl_FragColor = hyperbolic(gl_TexCoord[0].xy);
}